From b45d8c2e937231ab5f9d5a3094618e93caac09b9 Mon Sep 17 00:00:00 2001
From: Moshe Levi <moshele@mellanox.com>
Date: Thu, 6 Aug 2015 12:11:49 +0300
Subject: [PATCH] fix gfeature get size

---
 src/util/virnetdev.c |   76 +++++++++++++++++++++++++++++++++++++++++++++++--
 1 files changed, 72 insertions(+), 4 deletions(-)

diff --git a/src/util/virnetdev.c b/src/util/virnetdev.c
index 1e20789..f6a6733 100644
--- a/src/util/virnetdev.c
+++ b/src/util/virnetdev.c
@@ -89,8 +89,10 @@ VIR_LOG_INIT("util.netdev");
 
 #define RESOURCE_FILE_LEN 4096
 #if HAVE_DECL_ETHTOOL_GFEATURES
+#define ETH_GSTRING_LEN 32
 # define TX_UDP_TNL 25
-# define GFEATURES_SIZE 2
+# define DIV_ROUND_UP(n, d) (((n) + (d) - 1) / (d))
+# define FEATURE_BITS_TO_BLOCKS(n_bits)      DIV_ROUND_UP(n_bits, 32U)
 # define FEATURE_WORD(blocks, index, field)  ((blocks)[(index) / 32U].field)
 # define FEATURE_FIELD_FLAG(index)      (1U << (index) % 32U)
 # define FEATURE_BIT_IS_SET(blocks, index, field)        \
@@ -3110,6 +3112,69 @@ virNetDevGFeatureAvailable(const char *ifname, struct ethtool_gfeatures *cmd)
         ret = FEATURE_BIT_IS_SET(cmd->features, TX_UDP_TNL, active);
     return ret;
 }
+/**
+ * virNetDevGetGFeaturesSize
+ * This function return the number of gfeatures supported in
+ * the kernel
+ *
+ * @ifname: name of the interface
+ * @cmd: reference to a gfeatures ethtool command structure
+ *
+ * Returns gfeature size on success, and -1 on failure.
+ */
+ static int
+virNetDevGetGFeaturesSize(const char *ifname)
+{
+    struct {
+        struct ethtool_sset_info hdr;
+        __uint32_t buf[1];
+    }sset_info;
+    struct ethtool_gstrings *strings;
+    __uint32_t size;
+    int ret = -1;
+
+    sset_info.hdr.cmd = ETHTOOL_GSSET_INFO;
+    sset_info.hdr.reserved = 0;
+    sset_info.hdr.sset_mask = 1ULL << ETH_SS_FEATURES;
+
+    if (virNetDevSendEthtoolIoctl(ifname, &sset_info)){
+        return ret;
+    }
+    size = sset_info.hdr.sset_mask ? sset_info.hdr.data[0] : 0;
+
+    strings = calloc(1, sizeof(*strings) + size * ETH_GSTRING_LEN);
+    if (!strings){
+        return ret;
+    }
+
+    strings->cmd = ETHTOOL_GSTRINGS;
+    strings->string_set = ETH_SS_FEATURES;
+    strings->len = size;
+    if (size != 0 && virNetDevSendEthtoolIoctl(ifname, strings)) {
+        goto cleanup;
+    }
+
+    ret = 0;
+
+    if (strings) {
+        size = strings->len;
+    } else if (errno == EOPNOTSUPP || errno == EINVAL) {
+        /* Kernel doesn't support named features; not an error */
+        size = 0;
+    } else if (errno == EPERM) {
+        /* Kernel bug: ETHTOOL_GSSET_INFO was privileged.
+         * Work around it. */
+        size = 0;
+    }
+    size = FEATURE_BITS_TO_BLOCKS(size);
+
+ cleanup:
+    VIR_FREE(strings);
+    if (ret == 0)
+        ret = size;
+    return ret;
+}
+
 # endif
 
 
@@ -3189,9 +3254,12 @@ virNetDevGetFeatures(const char *ifname,
 
 # if HAVE_DECL_ETHTOOL_GFEATURES
     g_cmd.cmd = ETHTOOL_GFEATURES;
-    g_cmd.size = GFEATURES_SIZE;
-    if (virNetDevGFeatureAvailable(ifname, &g_cmd))
-        ignore_value(virBitmapSetBit(*out, VIR_NET_DEV_FEAT_TXUDPTNL));
+    g_cmd.size = virNetDevGetGFeaturesSize(ifname);
+    if (g_cmd.size == -1)
+        VIR_DEBUG("Failed to get GFeatures size in ifname %s", ifname);
+    else
+        if (virNetDevGFeatureAvailable(ifname, &g_cmd))
+            ignore_value(virBitmapSetBit(*out, VIR_NET_DEV_FEAT_TXUDPTNL));
 # endif
 
     if (virNetDevRDMAFeature(ifname, out) < 0)
-- 
1.7.1

